import register from "../util/register.js";

import m_assert_fact from "../util/assert.js";
import m_cfg_fact from "./config.js";
import m_data_fact from "./data.js";
import m_hud_fact from "./hud.js";
import m_print_fact from "./print.js";
import m_scenes_fact from "./scenes.js";
import m_subs_fact from "./subscene.js";
import m_trans_fact from "./transform.js";
import * as m_util from "./util.js";

/**
 * @name container
 * @namespace
 * @exports exports as container
 */
function Int_container(ns, exports) {

var m_assert  = m_assert_fact(ns);
var m_cfg     = m_cfg_fact(ns);
var m_data    = m_data_fact(ns);
var m_hud     = m_hud_fact(ns);
var m_print   = m_print_fact(ns);
var m_scenes  = m_scenes_fact(ns);
var m_subs    = m_subs_fact(ns);
var m_trans   = m_trans_fact(ns);

var cfg_def = m_cfg.defaults;
var cfg_lim = m_cfg.context_limits;

var _gl          = null;
var _canvas      = null;
var _canvas_hud  = null;
var _canvas_cont = null;
var _mouse_event_param = {clientX: 0, clientY: 0};

// NOTE: for optimization, to request canvas bounds only once a frame
var _offsets_updating_needed = false;

var _viewport_layout = {
    width: 0,   // device pixels
    height: 0,  // device pixels
    scale: 1,   // divice pixels/css pixels
    offset_top: 0,  // css pixels
    offset_left: 0  // css pixels
}

/**
 * Setup WebGL context
 * @param gl WebGL context
 */
exports.setup_context = function(gl) {
    _gl = gl;
}

exports.get_canvas = get_canvas;
function get_canvas() {
    return _canvas;
}

exports.get_canvas_hud = get_canvas_hud;
function get_canvas_hud() {
    return _canvas_hud;
}

exports.get_container = get_container;
function get_container() {
    return _canvas_cont;
}

exports.init = function(canvas, canvas_hud) {
    if (canvas && canvas.parentNode) {
        _canvas      = canvas;
        _canvas_hud  = canvas_hud;
        _canvas_cont = canvas.parentNode;
    } else
        m_assert.panic("canvas container is not available");
}

exports.insert_to_container = function(elem, stack_order) {

    stack_order = stack_order || "LAST";

    switch (stack_order) {
    case "FIRST":
        var cont_first_child = _canvas_cont.firstElementChild;

        _canvas_cont.insertBefore(elem, cont_first_child);

        break;
    case "JUST_BEFORE_CANVAS":
        _canvas_cont.insertBefore(elem, _canvas);

        break;
    case "JUST_AFTER_CANVAS":
        if (_canvas.nextElementSibling)
            _canvas_cont.insertBefore(elem, _canvas.nextElementSibling);
        else
            _canvas_cont.appendChild(elem, _canvas);

        break;
    case "LAST":
        _canvas_cont.appendChild(elem, _canvas);

        break;
    default:
        m_print.error(stack_order + " invalid stack order");

        break;
    }
}

exports.get_viewport_width = function() {
    return _viewport_layout.width;
}

exports.get_viewport_height = function() {
    return _viewport_layout.height;
}

exports.set_canvas_offsets = set_canvas_offsets;
function set_canvas_offsets(left, top) {
    _viewport_layout.offset_left = left;
    _viewport_layout.offset_top = top;
}

exports.update_canvas_offsets = update_canvas_offsets;
function update_canvas_offsets() {
    var boundaries = _canvas.getBoundingClientRect();
    set_canvas_offsets(boundaries.left, boundaries.top);
}

exports.setup_viewport_dim = setup_viewport_dim;
function setup_viewport_dim(width, height, scale) {
    _viewport_layout.width = width;
    _viewport_layout.height = height;
    _viewport_layout.scale = scale;
    update_canvas_offsets();
}

exports.force_offsets_updating = function() {
    _offsets_updating_needed = true;
}

function get_offset_top() {
    if (_offsets_updating_needed) {
        update_canvas_offsets();
        _offsets_updating_needed = false;
    }

    return _viewport_layout.offset_top;
}

function get_offset_left() {
    if (_offsets_updating_needed) {
        update_canvas_offsets();
        _offsets_updating_needed = false;
    }
    
    return _viewport_layout.offset_left;
}

exports.client_to_canvas_coords = function(client_x, client_y, dest) {
    return client_to_element_coords(client_x, client_y, _canvas, dest);
}

exports.get_coords_target_space = get_coords_target_space;
function get_coords_target_space(event, use_target_touches, dest) {

    if (event.constructor == MouseEvent) {
        dest[0] = event.offsetX;
        dest[1] = event.offsetY;
    } else if (event.constructor == TouchEvent) {
        if (event.type == "touchend")
            var touches = event.changedTouches;
        else
            var touches = use_target_touches ? event.targetTouches : event.touches;

        client_to_element_coords(touches[0].clientX, touches[0].clientY, 
                event.target, dest);
    } else {
        dest[0] = -1;
        dest[1] = -1;
    }

    return dest;
}

exports.client_to_element_coords = client_to_element_coords;
function client_to_element_coords(client_x, client_y, element, dest) {
    if (!cfg_def.ie11_edge_mouseoffset_hack) {
        // NOTE: hacky things.
        // Autoconvert client_x/client_y to offsetX/offsetY using custom MouseEvent
        _mouse_event_param.clientX = client_x;
        _mouse_event_param.clientY = client_y;

        // NOTE: needed in Chrome to work properly
        _mouse_event_param.view = window;
        var event = new MouseEvent("b4w_convert", _mouse_event_param);
        element.dispatchEvent(event);

        dest[0] = event.offsetX;
        dest[1] = event.offsetY;
    } else {
        // doesn't support complex CSS transformations, but still better 
        // than nothing
        var rect = element.getBoundingClientRect();
        dest[0] = client_x - rect.left;
        dest[1] = client_y - rect.top;
    }

    return dest;
}

exports.canvas_to_viewport_coords = function(canvas_x, canvas_y, dest, camera) {
    if (!dest)
        dest = new Float32Array(2);

    dest[0] = canvas_x * _viewport_layout.scale;
    dest[1] = canvas_y * _viewport_layout.scale;

    if (camera) {
        var camera_resolution_scale = camera.width / _viewport_layout.width;
        dest[0] *= camera_resolution_scale;
        dest[1] *= camera_resolution_scale;
    }

    return dest;
}

exports.viewport_to_canvas_coords = function(viewport_x, viewport_y, dest, camera) {
    if (!dest)
        dest = new Float32Array(2);
    
    dest[0] = viewport_x / _viewport_layout.scale;
    dest[1] = viewport_y / _viewport_layout.scale;

    if (camera) {
        var camera_resolution_scale = camera.width / _viewport_layout.width;
        dest[0] /= camera_resolution_scale;
        dest[1] /= camera_resolution_scale;
    }

    return dest;
}

exports.is_child = function(elem) {

    if (!elem || !elem.parentNode)
        return false;
    else if (elem.parentNode == _canvas_cont)
        return true;
    else
        return exports.is_child(elem.parentNode);
}

exports.is_hidpi = is_hidpi;
function is_hidpi() {
    if (cfg_def.allow_hidpi && window.devicePixelRatio >= 2)
        return true;
    return false;
}

function resize_css(canvas_webgl, canvas_hud, width, height) {
    canvas_webgl.style.width = width + "px";
    canvas_webgl.style.height = height + "px";

    if (canvas_hud) {
        canvas_hud.style.width = width + "px";
        canvas_hud.style.height = height + "px";
    }
}

exports.resize = resize;
function resize(width, height, update_canvas_css) {

    var canvas_webgl = get_canvas();
    var canvas_hud   = get_canvas_hud();
    if (update_canvas_css)
        resize_css(canvas_webgl, canvas_hud, width, height);

    if (canvas_hud) {
        // no HIDPI/resolution factor for HUD canvas
        canvas_hud.width  = width;
        canvas_hud.height = height;
        m_hud.update_dim();
    }

    var cw = Math.floor(width * cfg_def.canvas_resolution_factor);
    var ch = Math.floor(height * cfg_def.canvas_resolution_factor);

    canvas_webgl.width  = cw;
    canvas_webgl.height = ch;

    var width_limit = Math.min(_gl.drawingBufferWidth,
            cfg_lim.max_renderbuffer_size, cfg_lim.max_viewport_dims[0]);
    var height_limit = Math.min(_gl.drawingBufferHeight,
            cfg_lim.max_renderbuffer_size, cfg_lim.max_viewport_dims[1]);
    if (cw > width_limit || ch > height_limit) {
        m_print.warn("Canvas size exceeds platform limits, downscaling");

        var downscale = Math.min(width_limit / cw, width_limit / ch);

        cw *= downscale;
        ch *= downscale;
    }

    if (width)
        var scale = cw / width;
    else
        var scale = 1;

    canvas_webgl.width  = Math.floor(cw);
    canvas_webgl.height = Math.floor(ch);

    m_scenes.setup_dim(canvas_webgl.width, canvas_webgl.height, scale);

    // needed for frustum culling/constraints
    if (m_scenes.check_active())
        m_trans.update_transform(m_scenes.get_active()._camera);

    m_data.update_media_controls(canvas_webgl.width, canvas_webgl.height);
}

exports.resize_to_container = function(force) {
    var container = get_container();
    var canvas = get_canvas();

    var w = container.clientWidth;
    var h = container.clientHeight;

    if (force || w != canvas.clientWidth || h != canvas.clientHeight) {
        var old_width = canvas.width;
        var old_height = canvas.height;
        var canvas_webgl = get_canvas();
        var canvas_hud   = get_canvas_hud();
        resize_css(canvas_webgl, canvas_hud, w, h);

        // NOTE: in case of HMD splited screen, update canvas CSS width/height,
        // but keep canvas width/height unchangeable
        if (m_scenes.check_active()) {
            var active_scene = m_scenes.get_active();
            var subs_stereo = m_scenes.get_subs(active_scene, m_subs.STEREO);
            if (subs_stereo && subs_stereo.enable_hmd_stereo) {
                // restore canvas.width/canvas.height
                resize(old_width, old_height, false)
                return;
            }
        }
        resize(w, h, false)
    }
}

exports.reset = function() {
    _gl          = null;
    _canvas      = null;
    _canvas_hud  = null;
    _canvas_cont = null;
}

}

var int_container_factory = register("__container", Int_container);

export default int_container_factory;
